var $ = require('./utils');

var path = require('path'),
    Promise = require('es6-promise'),
    cliWidth = require('cli-width');

var logger = require('./logger');

var prefix = 'data:application/json;charset=utf-8;base64,';

function toUrl(sourceMap) {
  return prefix + new Buffer(JSON.stringify(sourceMap)).toString('base64');
}

module.exports = function(tarima, files, cb) {
  var _width = cliWidth();

  var start = +new Date();

  var data = [];
  var tasks = [];

  var cache = this.cache,
      state = this.state,
      options = this.opts;

  var dist = this.dist,
      match = this.match,
      dispatch = this.dispatch;

  options.bundleOptions.cache = cache.all() || {};

  // built-in helpers
  options.bundleOptions.helpers.import = function(id) {
    return $.read(path.join(options.dest, id));
  };

  options.bundleOptions.helpers.include = function() {
    return Array.prototype.slice.call(arguments)
      .map(function(src) {
        if (src.indexOf('.css') > -1) {
          return '<link rel="stylesheet" href="' + src + '">';
        }

        if (src.indexOf('.js') > -1) {
          return '<script src="' + src + '"></script>';
        }
      })
      .join('\n');
  };

  options.bundleOptions.helpers.resources = function() {
    return (options.bundleOptions.resources || []).join('\n');
  };

  if ($.isFile(this.opts.rollupFile)) {
    $.merge(options.bundleOptions.rollup, require(path.resolve(this.opts.rollupFile)));
  }

  function tarimaOptions() {
    var params = {};

    params.cwd = options.cwd;

    for (var key in options.bundleOptions) {
      params[key] = options.bundleOptions[key];
    }

    return params;
  }

  var fixedBundle = $.toArray(options.bundle);

  var isBundle = function() {
    return false;
  };

  var _cache;
  var _bundle;

  if (options.bundleOptions.entryCache) {
    _cache = {};
  }

  for (var i = 0, c = fixedBundle.length; i < c; i++) {
    if (fixedBundle[i] === true) {
      fixedBundle = [];

      isBundle = function() {
        return true;
      };

      break;
    }

    if (!fixedBundle[i]) {
      fixedBundle.splice(i, 1);
    }
  }

  if (fixedBundle.length) {
    isBundle = $.makeFilter(true, fixedBundle);
  }

  // custom events
  var onWrite = this.emit.bind(null, 'write'),
      onDelete = this.emit.bind(null, 'delete');

  function ensureRename(view) {
    if (typeof options.rename === 'function') {
      options.rename(view);
    }
  }

  function ensureWrite(view, index) {
    onWrite(view, index);

    if (options.bundleOptions.compileDebug && view.sourceMap) {
      // TODO: normalize sourceMap

      delete view.sourceMap.sourceRoot;
      delete view.sourceMap.sourcesContent;

      view.sourceMap.file = view.dest;
      view.sourceMap.sources = (view.sourceMap.sources || [view.src])
        .map(function(src) {
          return path.relative(options.cwd, src);
        });

      if (!view.sourceMap.sources[0]) {
        view.sourceMap.sources[0] = view.src;
      }

      view.sourceMap.sourcesContent = view.sourceMap.sources
        .map(function(src) {
          return $.read(src).toString();
        });

      if (options.bundleOptions.sourceMapFile === true) {
        $.write(view.dest + '.map', JSON.stringify(view.sourceMap));
      }

      var prefix = view.dest.indexOf('.js') > -1 ? '//' : '/*';
      var suffix = view.dest.indexOf('.js') > -1 ? '' : ' */';

      view.output += '\n' + prefix + '# sourceMappingURL=' +
        (options.bundleOptions.sourceMapFile ? path.basename(view.dest) + '.map' : toUrl(view.sourceMap))
        + suffix;
    }

    $.write(view.dest, view.output);
  }

  function ensureWait(src) {
    var _file = path.relative(options.cwd, src);
    var _max = Math.max(42, _width) - 15;

    if (_file.length > _max) {
      _file = _file.substr(0, 10) + '...' + _file.substr(-(_max - 13));
    }

    logger.printf('\r{wait.yellow|%s}', _file);
    logger.erase('lineRight');
    logger.printf('\r');
  }

  function dest(id, ext) {
    return path.relative(options.cwd, path.join(options.dest, ext ? id.replace(/\.[\w.]+$/, '.' + ext) : id));
  }

  function sync(id, resolve) {
    var entry = cache.get(id) || {};

    entry.mtime = $.mtime(id);

    delete entry.dirty;

    if (resolve) {
      resolve(entry);
    }

    cache.set(id, entry);
  }

  function copy(src) {
    return function(next) {
      var out = [];

      src.forEach(function(target) {
        var entry = cache.get(target.src);

        if ((entry && entry.deleted) || !$.exists(target.src)) {
          target.type = 'delete';
          dist(target);
          return;
        }

        var _start = new Date();

        ensureWait(target.src);

        logger.status(_start, 'copy', target, function() {
          sync(target.src);

          out.push(target.dest);
          entry.dest = target.dest;

          $.copy(target.src, target.dest);
        });
      });

      data.push(out);

      next();
    };
  }

  function track(src, sub) {
    sync(src, function(entry) {
      entry.deps = entry.deps || [];

      (sub || []).forEach(function(dep) {
        sync(dep, function(_entry) {
          _entry.deps = _entry.deps || [];

          if (_entry.deps.indexOf(src) === -1) {
            _entry.deps.push(src);
          }
        });
      });
    });
  }

  function append(src, next) {
    var entry = cache.get(src);

    if (entry.dirty === false) {
      delete entry.dirty;
    }

    if (entry.dirty) {
      next(src);
    }
  }

  function compile(src) {
    var entry = cache.get(src) || {};

    // override
    entry.main = true;

    if (entry.deleted) {
      dist({
        type: 'delete',
        src: src,
        dest: entry.dest
      });

      cache.rm(src);
      onDelete(src, entry);
    } else {
      var opts = tarimaOptions();

      try {
        var partial = tarima.load(path.join(options.cwd, src), opts);
      } catch (e) {
        return cb(e);
      }

      tasks.push(function(next) {
        ensureWait(src);

        var _start = new Date();
        var prefix = 'render';

        if (partial.params.isTemplate || partial.params.data._bundle || isBundle(src)) {
          opts._bundle = _bundle;
          opts._cache = _cache;

          prefix = partial.params.data._bundler || opts.bundler
            || (partial.params.isScript ? 'rollup' : 'bundle');

          partial = tarima.bundle(partial, opts);
        }

        partial.render(function(err, output) {
          if (err) {
            return cb(err);
          }

          // cached for later
          if (options.bundleOptions.bundleCache && prefix === 'rollup') {
            _bundle = output._bundle || _bundle;
          }

          var file = path.relative(options.cwd, output.filename),
              target = dest(file, output.extension);

          var index = track.bind(null, file);

          var result = {
            src: file,
            dest: target
          };

          ensureRename(result);

          logger.status(_start, prefix, result, function() {
            result.sourceMap = output.sourceMap;
            result.output = output.source;

            // TODO: only track partials (?)
            var fixedDeps = entry.deps || [];

            output.deps.forEach(function(id) {
              if (id.indexOf(options.cwd) !== 0) {
                return;
              }

              var dep = path.relative(options.cwd, id);

              if ((file.split('/')[0] === dep.split('/')[0])
                && fixedDeps.indexOf(dep) === -1) {
                fixedDeps.push(dep);
              }
            });

            index(fixedDeps);
            ensureWrite(result, index);

            cache.set(file, 'deps', fixedDeps);
            cache.set(file, 'dest', result.dest);

            delete result.output;
          });

          data.push(result.dest);

          next();
        });
      });
    }
  }

  var seen = {};
  var unknown = [];

  files.forEach(function(src) {
    var entry = cache.get(src);

    if (!entry) {
      // required reference
      cache.set(src, { dirty: true });
    }

    if (!tarima.support.isSupported(src)) {
      return append(src, function(id) {
        if (match(src)) {
          seen[id] = 1;
          unknown.push(id);
        }
      });
    }

    if (_cache && entry && entry.dirty) {
      delete _cache[src];
    }

    if (!seen[src] && match(src)) {
      seen[src] = true;
      append(src, compile);
    }
  });

  if (unknown.length) {
    dispatch(unknown.map(function(file) {
      var _target = {
        src: file,
        dest: dest(file)
      };

      ensureRename(_target);

      return _target;
    }), function(src, run) {
      (src.length ? [copy(src)] : []).concat(run)
        .forEach(function(cb) {
          tasks.push(cb);
        });
    });
  }

  function _end(err) {
    // reset
    _cache = null;
    _bundle = null;

    cb(err, {
      cache: cache.all(),
      input: Object.keys(seen),
      output: $.flatten(data),
      elapsed: $.timeDiff(start)
    });
  }

  tasks.reduce(function(a, b) {
    return a.then(function() {
      if (state('abort')) {
        return;
      }

      return new Promise(function(resolve, reject) {
        b(function(err) {
          if (err) {
            reject(err);
          } else {
            resolve();
          }
        });
      });
    });
  }, Promise.resolve())
    .then(function() {
      if (state('abort')) {
        logger.printf('\r{gray.diff|The build was stopped}');
        logger.erase('lineRight');
        logger.printf('\n');
      }

      _end();
    })
    .catch(function(e) { _end(e); } );
};
