var tarima = require('tarima');

var $ = require('./utils');

var path = require('path'),
    Promise = require('es6-promise'),
    micromatch = require('micromatch');

var logger = require('./logger'),
    readFiles = require('./read'),
    compileFiles = require('./compile');

var plugableSupportAPI = require('./hooks'),
    cacheableSupportAPI = require('./caching');

function makeReplacement(obj, test, rename) {
  return function(value) {
    var dir = obj.dest || obj.cwd;
    var rel = value ? path.relative(dir, value) : path.relative(obj.cwd, dir);
    var ok = test(rel);

    if (ok) {
      var ext = path.extname(rel);

      // support for dynamic path slicing and rename strategy
      return path.join(dir, (rename || ok).replace(/\{filepath(?:\/(.+?))?\}/, function(_, match) {
          var parts = path.dirname(rel).split('/');
          var keys = match ? match.split('/') : [];

          var h = 0;
          var j = 0;
          var test = [];

          while (true) {
            var a = parts[j], b = keys[h];

            if (typeof a === 'undefined' && b) {
              break;
            }

            if (typeof b === 'undefined' && typeof a === 'undefined') {
              break;
            }

            if (/^\d+$/.test(b)) {
              parts.splice(j, +b);
              j = 0;
              h++;
              continue;
            }

            if (a === b) {
              h++;
              parts[j] = keys[h];
              h++;
              j++;
              continue;
            }

            test.push(a);
            j++;
          }

          return test.join('/');
        })
        .replace('{filename}', path.basename(rel, ext))
        .replace('{extname}', ext.substr(1)));
    }
  };
}

function complexFactor(test) {
  return test.split('/').length * test.split('**').length;
}

module.exports = function(options, done) {
  if (!Array.isArray(options.bundle)) {
    options.bundle = options.bundle === true ? ['**'] : options.bundle ? [options.bundle] : [];
  }

  // resolve all relative paths
  ['dest', 'public', 'reload', 'cacheFile', 'rollupFile'].forEach(function(subpath) {
    if (options[subpath]) {
      options[subpath] = Array.isArray(options[subpath]) ?
        options[subpath].map(function(subdir) {
          return path.resolve(options.cwd, subdir);
        }) : path.resolve(options.cwd, options[subpath]);
    }
  });

  var filters = Array.isArray(options.filter) ? options.filter : ['**'];

  $.toArray(options.watch).forEach(function(value) {
    if (typeof value === 'string') {
      filters.push('!' + value + '/**');
    }
  });

  var context = plugableSupportAPI(logger, options);

  // normalize options
  options.port = options.port || process.env.PORT || 3000;

  options.ignore = options.ignore || [];
  options.locals = options.locals || {};
  options.rename = options.rename || [];

  // devPlugins works only on watch/server mode
  options.plugins = options.plugins || [];
  options.devPlugins = options.devPlugins || [];

  // internally used
  context.cache = cacheableSupportAPI(options.cacheFile);
  context.match = $.makeFilter(false, filters);

  options.pluginOptions = options.pluginOptions || {};
  options.bundleOptions = options.bundleOptions || {};

  options.bundleOptions.locals = options.bundleOptions.locals || {};
  options.bundleOptions.globals = options.bundleOptions.globals || {};

  options.bundleOptions.rollup = options.bundleOptions.rollup || {};
  options.bundleOptions.helpers = options.bundleOptions.helpers || {};
  options.bundleOptions.resources = options.bundleOptions.resources || [];

  function die(error) {
    done.call(context, Array.isArray(error) ? error.map(function(err) {
      return err.message || err.toString();
    }).join('\n') : error);
  }

  if (!$.exists(options.dest)) {
    options.force = true;
  }

  var src = [];

  var plugs = Promise.all((options.plugins || [])
    // conditionally load devPlugins
    .concat((options.watch || options.server) ? options.devPlugins || [] : [])
      .map(function(file) {
        var testFile = path.resolve(file);

        return require($.isFile(testFile) ? testFile : file);
      })
      .sort(function(a, b) {
        return b.length - a.length;
      })
      .map(function(cb) {
        if (!cb) {
          return;
        }

        return new Promise(function(resolve, reject) {
          function next(e) {
            if (e) {
              reject(e);
            } else {
              resolve();
            }
          }

          // ES6 interop
          cb = cb.default || cb;

          if (cb.length) {
            cb.call(context, next);
          } else {
            cb.call(context);
            next();
          }
        });
      }));

  var opts = plugs.then(function() {
    if (options.ignoreFiles) {
      options.ignoreFiles.forEach(function(ifile) {
        if ($.isFile(ifile)) {
          var lines = $.read(ifile).toString().split('\n');

          lines.forEach(function(line) {
            if (line.length && (line[0] !== '#') && (options.ignore.indexOf(line) === -1)) {
              var offset = line.indexOf('/');

              if (offset === -1) {
                options.ignore.push('**/' + line);

                if (line.indexOf('.') === -1 && line.indexOf('*') === -1) {
                  options.ignore.push('**/' + line + '/**');
                  options.ignore.push(line + '/**');
                }

                options.ignore.push(line);
                return;
              }

              if (offset === 0) {
                line = line.substring(1);
              }

              if (line.charAt(line.length - 1) === '/') {
                options.ignore.push(line.slice(0, -1));
                options.ignore.push(line + '**');
              } else {
                options.ignore.push(line);
              }
            }
          });
        }
      });
    }

    if (options.rename) {
      var replaceOpts = $.toArray(options.rename).map(function(test) {
        if (typeof test === 'function') {
          return { cb: makeReplacement(options, test) };
        }

        if (typeof test === 'string') {
          test = test.split(':');
        }

        if (Array.isArray(test)) {
          var re = {},
              isRe = test[0] instanceof RegExp;

          re[isRe ? 're' : 'str'] = test[0];

          if (!isRe) {
            test[0] = micromatch.makeRe(test[0]);
          }

          re.cb = makeReplacement(options, RegExp.prototype.test.bind(test[0]), test[1]);

          return re;
        }
      });

      // TODO: organize helpers...
      replaceOpts.sort(function(a, b) {
        if (a.cb) {
          return b.cb ? 0 : b.re ? 1 : -1;
        }

        if (a.re) {
          return b.cb ? 1 : b.re ? 0 : -1;
        }

        return a.str && b.str ? complexFactor(b.str) - complexFactor(a.str) : 1;
      });

      options.rename = function(view) {
        for (var key in replaceOpts) {
          view.dest = replaceOpts[key].cb(view.dest) || view.dest;
        }

        view.dest = path.relative(options.cwd, view.dest);
      };
    }
  });

  var _state = 'pending';

  context.state = function(value) {
    return _state === value;
  };

  var close;

  function end(err, result) {
    if (close) {
      close.call(context, err, result);
      close = null;
    }

    context.cache.save();
    context.emit('end', err, result);

    done.call(context, err, result);

    if (typeof options.reload === 'string') {
      logger.getLogger()
        .INFO('{log.gray|Running %s}', options.reload);

      options.reload = require(options.reload);
    }

    if (typeof options.reload === 'function') {
      close = options.reload;
    }

    setTimeout(function() {
      _state = 'pending';
    });
  }

  function build(err) {
    if (!err) {
      compileFiles.call(context, tarima, src, end);
      _state = 'working';
    } else {
      end(err);
    }
  }

  try {
    opts
    .catch(die)
    .then(function() {
      readFiles.call(context, function(err2, files) {
        if (_state === 'working') {
          _state = 'abort';
        }

        if (_state === 'pending') {
          src = files;
          build(err2);
        }
      });
    });
  } catch (e) {
    die(e);
  }
};
