var $ = require('./utils');

var clc = require('chalk');

var write = process.stdout.write.bind(process.stdout);

var levels = [0, 'info', 'debug', 'verbose'],
    current = 0;

var symbols = {
  ok: '✔',
  err: '✗',
  log: '—',
  diff: '≠',
  warn: '⚠',
  info: 'ℹ',
  hint: '›',
  wait: '↺'
};

var styles = {
  delete: 'ok.red',
  error: 'err.red'
};

var erase = {
  lineLeft: '\x1b[1K',
  lineRight: '\x1b[K'
};

function style(message) {
  return message.replace(/\{([.\w]+)\|(.+?)\}(?=\s|$)/g, function(_, format, message) {
    var colorized = clc,
        segments = format.split('.');

    var depth = 8;

    while (segments.length) {
      var key = segments.shift();

      if (key === 'pad') {
        message = (new Array(depth - message.length + 1)).join(' ') + message;
        continue;
      }

      if (symbols[key]) {
        message = symbols[key] + ' ' + message;
        continue;
      }

      if (!colorized[key]) {
        break;
      }

      colorized = colorized[key];
    }

    if (typeof colorized === 'function') {
      return colorized(message);
    }

    return message;
  }).replace(/`(.+?)`/g, function(expression) {
    return clc.yellow(expression);
  });
}

function puts(message) {
  var args = Array.prototype.slice.call(arguments, 1);

  return String(message).replace(/%s/g, function() {
    return args.shift();
  });
}

function log(allowed) {
  return function() {
    if ((allowed - 1) < current) {
      write('\r' + style(puts.apply(null, arguments)) + '\n');
    }

    return this;
  };
}

module.exports = {
  erase: function(type, suffix) {
    write(erase[type] + (suffix || ''));
  },
  status: function(start, type, out, cb) {
    if (!(start instanceof Date)) {
      cb = out;
      out = type;
      type = start;
      start = null;
    }

    if (typeof type === 'object') {
      cb = out;
      out = type;
      type = out.type || 'unknown';
    }

    var ok = this.isEnabled();

    var src = out.src || out,
        dest = out.dest || out;

    if (src) {
      if (Array.isArray(src)) {
        src = '[' + src.length + ' file' + (src.length !== 1 ? 's' : '') + ']';
      } else {
        src = out.src || out;
      }
    }

    var err;

    try {
      if (cb) {
        cb();
      }
    } catch (e) {
      err = e;
    }

    if (ok) {
      var diff = $.timeDiff(start);
      var base = 'gray';

      // Ns (seconds)
      if (diff.indexOf('ms') === -1) {
        base = 'yellow';

        // TODO: configure threshold?
        if (parseFloat(diff) > 2.0) {
          base = 'red';
        }
      }

      var ms = start ? ' {' + base + '|+' + diff + '}' : '';

      if (err) {
        this.printf('\r  {pad.gray|%s} {err.red|%s}%s', type, src || dest, ms);
        this.writeln(err);
      } else {
        this.printf('\r  {pad.gray|%s} {%s|%s}%s', type, styles[type] || 'ok.green', dest, ms);
      }

      write(erase.lineRight + '\n');
    }
  },
  printf: function() {
    write(style(puts.apply(null, arguments)));
  },
  writeln: function() {
    write('\r' + puts.apply(null, arguments) + '\n');
  },
  setLevel: function(type) {
    current = typeof type !== 'number' ? levels.indexOf(type) : type;
  },
  getLogger: function() {
    return {
      INFO: log(1),
      DEBUG: log(2),
      VERBOSE: log(3)
    };
  },
  isEnabled: function() {
    return current > 0;
  },
  isDebug: function() {
    return current > 1;
  }
};
