var fs = require('fs-extra'),
    path = require('path'),
    clc = require('chalk'),
    prettyMs = require('pretty-ms'),
    micromatch = require('micromatch');

var notifier;

try {
  notifier = require('node-notifier');
} catch (e) { /* nothing */ }

function env(value) {
  if (typeof value === 'string') {
    return value.replace(/\$\{(\w+)\}/g, function(match, key) {
      return process.env[key] || match;
    });
  }

  return value;
}

function merge(target, source) {
  for (var key in source) {
    var value = source[key];

    if (Array.isArray(value) && (!target[key] || Array.isArray(target[key]))) {
      if (!target[key]) {
        target[key] = [];
      }

      value.forEach(function(item) {
        var fixedItem = env(item);

        if (target[key].indexOf(fixedItem) === -1) {
          target[key].push(fixedItem);
        }
      });
    } else if (value !== null && typeof value === 'object') {
      merge(target[key] || (target[key] = {}), value);
    } else if (typeof value !== 'undefined') {
      target[key] = env(value);
    }
  }
}

function flatten(items) {
  var out = [];

  items.forEach(function(set) {
    if (Array.isArray(set)) {
      Array.prototype.push.apply(out, flatten(set));
    } else if (set) {
      out.push(set);
    }
  });

  return out;
}

function isDir(filepath) {
  try {
    return fs.statSync(filepath).isDirectory();
  } catch (e) {
    // noop
  }
}

function isFile(filepath) {
  try {
    return fs.statSync(filepath).isFile();
  } catch (e) {
    // noop
  }
}

function exists(filepath) {
  return isDir(filepath) || isFile(filepath);
}

function timeDiff(start) {
  return prettyMs((new Date()) - start);
}

function mtime(filepath) {
  return exists(filepath) ? +fs.statSync(filepath).mtime : null;
}

function toArray(obj) {
  if (!obj) {
    return [];
  }

  return !Array.isArray(obj) ? [obj] : obj;
}

function notify(message, title, icon) {
  var noticeObj = {
    title: title,
    message: message
  };

  if (exists(icon)) {
    noticeObj.icon = path.resolve(icon);
  }

  if (notifier) {
    notifier.notify(noticeObj);
  }
}

function makeFilter(any, filters) {
  filters = filters.map(function(filter) {
    return micromatch.matcher(filter, {
      dot: true
    });
  });

  // micromatch.filter() didn't work as expected
  return function(filepath) {
    var length = filters.length;
    var res = false;
    var pass = 0;

    while (length--) {
      if (filters[length](filepath)) {
        if (any) {
          res = true;
          break;
        }

        pass++;
      }
    }

    return res || pass === filters.length;
  };
}

module.exports = {
  makeFilter: makeFilter,
  notify: notify,
  merge: merge,
  exists: exists,
  isDir: isDir,
  isFile: isFile,
  copy: fs.copySync,
  unlink: fs.unlinkSync,
  readJSON: fs.readJsonSync,
  writeJSON: fs.outputJsonSync,
  timeDiff: timeDiff,
  mtime: mtime,
  write: fs.outputFileSync,
  read: fs.readFileSync,
  toArray: toArray,
  flatten: flatten,
  color: clc
};
